from click import ClickException
import os

import ray.projects


# Pathnames specific to Ray's project directory structure.
PROJECT_ID_BASENAME = 'project-id'
RAY_PROJECT_DIRECTORY = '.rayproject'


def get_project_id(project_dir):
    """
    Args:
        project_dir: Project root directory.

    Returns:
        The ID of the associated Project in the database.

    Raises:
        ValueError: If the current project directory does not contain a project ID.
    """
    project_id_filename = os.path.join(project_dir, RAY_PROJECT_DIRECTORY, PROJECT_ID_BASENAME)
    if not os.path.isfile(project_id_filename):
        raise ClickException("Ray project in {} not registered yet. Did you run 'any project create'?".format(project_dir))
    with open(project_id_filename, 'r') as f:
        project_id = f.read()
    try:
        project_id = int(project_id)
    except:
        raise ClickException("{} does not contain a valid project ID".format(project_id_filename))
    return project_id



import boto3
import click
import datetime
import json
import os
import requests

import anyscale.conf
from anyscale.common import ENDPOINTS


def confirm(msg, yes):
    return None if yes else click.confirm(msg, abort=True)

def get_endpoint(endpoint):
    route = ENDPOINTS[endpoint]
    return "{}{}".format(anyscale.conf.APPLICATION_URL, route)

def load_credentials():
    path = os.path.expanduser("~/.anyscale/credentials.json")
    if not os.path.exists(path):
        raise click.ClickException("Please create a token by going to {}/cli".format(anyscale.conf.APPLICATION_URL))
    with open(path) as f:
        try:
            credentials = json.load(f)
        except json.JSONDecodeError as e:
            raise click.ClickException(e)
    if "cli_token" not in credentials:
        raise click.ClickException("Credential file not valid, please regenerate it.")
    return credentials["cli_token"]

def send_json_request(endpoint, json_args, post=False):
    if anyscale.conf.CLI_TOKEN is None:
        anyscale.conf.CLI_TOKEN = load_credentials()

    url = get_endpoint(endpoint)
    cookies = {"cli_token": anyscale.conf.CLI_TOKEN}
    try:
        if post:
            resp = requests.post(url, json=json_args, cookies=cookies)
        else:
            resp = requests.get(url, params=json_args, cookies=cookies)
    except requests.exceptions.ConnectionError as e:
        raise click.ClickException("Failed to connect to anyscale server at {}".format(url))

    if not resp.ok:
        raise click.ClickException("{}: {}.".format(resp.status_code, resp.text))
    json_resp = resp.json()
    if "error" in json_resp:
        raise click.ClickException("Error: {}".format(json_resp["error"]))

    return json_resp

def serialize_datetime(d):
    # Make sure that overwriting the tzinfo in the line below is fine.
    # Note that we have to properly convert the timezone if one is already specified.
    # This can be done with the .astimezone method.
    assert d.tzinfo is None
    return d.replace(tzinfo=datetime.timezone.utc).isoformat()

def deserialize_datetime(s):
    return datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%f+00:00")

def humanize_timestamp(timestamp):
    delta = datetime.datetime.utcnow() - timestamp
    offset = delta.seconds + (delta.days * 60 * 60 * 24)
    delta_s = int(offset % 60)
    offset /= 60
    delta_m = int(offset % 60)
    offset /= 60
    delta_h = int(offset % 24)
    offset /= 24
    delta_d = int(offset)

    if delta_d >= 1:
        return "{} day{} ago".format(delta_d, "s" if delta_d > 1 else "")
    if delta_h > 0:
        return "{} hour{} ago".format(delta_h, "s" if delta_h > 1 else "")
    if delta_m > 0:
        return "{} minute{} ago".format(delta_m, "s" if delta_m > 1 else "")
    else:
        return "{} second{} ago".format(delta_s, "s" if delta_s > 1 else "")



import json
import logging
import requests
import os
import shutil
import subprocess
import time

from git import Git
import ray.ray_constants
import yaml

from anyscale.project import get_project_id
from anyscale.common import (COMMIT_PATCH_BASENAME)
from anyscale.util import (confirm, send_json_request)


logging.basicConfig(format=ray.ray_constants.LOGGER_FORMAT)
logger = logging.getLogger(__file__)

# A temporary directory to download snapshots to.
TEMP_SNAPSHOT_DIRECTORY = '/tmp/anyscale-snapshot-{}'

def copy_file(to_s3, source, target, download):
    """Copy a file.

    The file source or target may be on S3.

    Args:
        to_s3 (bool): If this is True, then copy to/from S3, else the local
            disk. If this is True, then the file source or target will be a
            presigned URL to which GET or POST HTTP requests can be sent.
        source (str or S3 URL): Source file local pathname or S3 GET URL. If
            this is an S3 URL, target is assumed to be a local pathname.
        target (str or S3 URL): Target file local pathname or S3 URL with POST
            credentials. If this is an S3 URL, source is assumed to be a local
            pathname.
        download (bool): If this is True, then this will upload from source to
            target, else this will download.
    """
    try:
        if to_s3:
            if download:
                with open(target, 'wb') as f:
                    response = requests.get(source)
                    for block in response.iter_content(1024):
                        f.write(block)
            else:
                with open(source, 'rb') as f:
                    files = {'file': ('object', f)}
                    resp = requests.post(target['url'], data=target['fields'], files=files)
                    assert resp.ok, resp.text
        else:
            shutil.copyfile(source, target)
    except (OSError, AssertionError) as e:
        logger.warn("Failed to copy file %s , aborting", source)
        raise e


def walk_files(roots, exclude_files=None):
    """Helper function to get all snapshot files with a recursive walk.

    Args:
        roots: Top-level files to save. These may be directories.
        exclude_files: A list of files to skip during copy.

    Returns:
        (list, list): Tuple of (directories, files) seen when recursively
            walking the roots.
    """
    if exclude_files is None:
        exclude_files = []
    directories = []
    files = []
    for root in roots:
        if root in exclude_files:
            continue
        if os.path.isdir(root):
            directories.append(root)
            contents = [os.path.join(root, file) for file in os.listdir(root)]
            subdirectories, subfiles = walk_files(contents, exclude_files=exclude_files)
            directories += subdirectories
            files += subfiles
        else:
            files.append(root)

    return directories, files


def get_snapshot_files(project_definition):
    """Get all file metadata needed to create a snapshot.

    Args:
        project_definition: Project defininition.

    Returns:
        dict describing all file metadata needed for the snapshot.
    """
    project_dir = project_definition.root

    # Get the output files to save with the snapshot.
    output_files = []
    missing_output_files = []
    for output_file in project_definition.config.get('output_files', []):
        if os.path.exists(os.path.join(project_dir, output_file)):
            output_files.append(output_file)
        else:
            missing_output_files.append(output_file)

    # Get the input files to save with the snapshot. By default, this is all
    # files in the project directory that are not tracked by git and are not
    # explicitly specified by the user to be an output file in project.yaml.
    input_files = []
    commit_hash = None
    commit_patch_location = None
    if project_definition.git_repo() is not None:
        # Get and save the current git repo state.
        git = Git('.')
        try:
            commit_hash = git.rev_parse('HEAD')
        except:
            pass

        # We succeeded in getting a git hash. Get the rest of the git state.
        if commit_hash is not None:
            # Find all files in the project that are not already tracked by git.
            for untracked_file in git.ls_files(project_dir, exclude_standard=True, others=True, directory=True).split('\n'):
                # Save the pathname relative to the project so that we can later
                # recreate the directory structure inside the project.
                untracked_file = os.path.relpath(untracked_file, project_dir)
                input_files.append(untracked_file)
            commit_hash = git.rev_parse('HEAD')
            # Save the git diff, if any.
            diff = git.diff()
            if diff:
                filename = "{}.{}".format(COMMIT_PATCH_BASENAME, time.time())
                commit_patch_location = os.path.join("/tmp", filename)
                with open(commit_patch_location, 'w+') as f:
                    f.write(diff)
                    f.write('\n')
    else:
        for file in os.listdir(project_dir):
            input_files.append(file)

    # Confirm the files to save with the snapshot.
    if input_files or output_files or missing_output_files:
        print("Creating a new snapshot for this session, containing the files:")

        if input_files:
            print("Input files:")
            for file in input_files:
                print(" ", file)
        if output_files:
            print("Output files:")
            for file in output_files:
                print(" ", file)
        if missing_output_files:
            print("Missing output files:")
            for file in missing_output_files:
                print(" ", file)

    # Snapshot all found input and output files to disk or S3.
    index = 0
    input_mappings = {}
    output_mappings = {}
    # Recursively get all input and output files.
    input_directories, input_files = walk_files(input_files, exclude_files=output_files)
    output_directories, output_files = walk_files(output_files)
    return {
            "commit_hash": commit_hash,
            "commit_patch_location": commit_patch_location,
            "input_directories": input_directories,
            "input_files": input_files,
            "output_directories": output_directories,
            "output_files": output_files,
            "missing_output_files": missing_output_files,
            }


def create_snapshot(project_definition, yes,
        description=None, local=False):
    """Create a snapshot of a project.

    Args:
        project_definition: Project definition.
        yes: Don't ask for confirmation.
        description: An optional description of the snapshot.
        local: Whether the snapshot should be of a live session or
            the local project directory state.

    Raises:
        ValueError: If the current project directory does not match the project
            metadata entry in the database.
        Exception: If saving the snapshot files fails.
    """
    if not local:
        raise NotImplementedError("Snapshots of a live session not currently supported.")

    # Find and validate the current project ID.
    project_dir = project_definition.root
    project_id = get_project_id(project_dir)

    files = get_snapshot_files(project_definition)
    with open(project_definition.cluster_yaml()) as f:
        cluster_config = yaml.safe_load(f)
    resp = send_json_request("snapshot_create", {
        "project_id": project_id,
        "project_config": json.dumps(project_definition.config),
        "cluster_config": json.dumps(cluster_config),
        "description": description if description else "",
        "files": files,
        }, post=True)
    snapshot_uuid = resp["uuid"]
    snapshot_to_s3 = resp["snapshot_to_s3"]
    mappings = resp["locations"]

    try:
        for source, target in mappings.items():
            copy_file(snapshot_to_s3, source, target, download=False)
    except (OSError, AssertionError) as e:
        # Tell the server to delete the snapshot.
        send_json_request("snapshot_delete", {
            "snapshot_uuids": [snapshot_uuid],
            }, post=True)
        raise e
    return snapshot_uuid


def delete_snapshot(project_dir, uuid, yes):
    """Delete the snapshot(s) with the given name.

    Delete the snapshot data from disk and the metadata from the database.

    Args:
        project_dir: Project root directory.
        uuid: The UUID of the snapshot to delete.
        yes: Don't ask for confirmation.

    Raises:
        ValueError: If the current project directory does not match the project
            metadata entry in the database.
    """
    # Find and validate the current project ID.
    project_id = get_project_id(project_dir)

    # Get the snapshots with the requested name.
    resp = send_json_request("snapshot_list", {
        "project_id": project_id,
        "snapshot_uuid": uuid,
        })
    snapshots = resp["snapshots"]
    if snapshots:
        snapshots_str = '\n'.join(snapshot["uuid"] for snapshot in snapshots)
        confirm("Delete snapshots?\n{}".format(snapshots_str), yes)
        snapshot_uuids = [snapshot["uuid"] for snapshot in snapshots]
        send_json_request("snapshot_delete", {
            "snapshot_uuids": snapshot_uuids,
            }, post=True)
    else:
        logger.warn("No snapshots found with UUID %s", uuid)

def describe_snapshot(uuid):
    resp = send_json_request("snapshot_describe", {
        "snapshot_uuid": uuid,
    })
    return resp

def list_snapshots(project_dir):
    """List all snapshots associated with the given project.

    Args:
        project_dir: Project root directory.

    Returns:
        List of Snapshots for the current project.

    Raises:
        ValueError: If the current project directory does not match the project
            metadata entry in the database.
    """
    # Find and validate the current project ID.
    project_id = get_project_id(project_dir)
    resp = send_json_request("snapshot_list", {
        "project_id": project_id,
        })
    snapshots = resp["snapshots"]
    return [snapshot["uuid"] for snapshot in snapshots]


def get_snapshot_uuid(project_dir, snapshot_uuid):
    """Get a snapshot of the given project with the given name.

    Args:
        project_id: The ID of the project.
        snapshot_name: The name of the snapshot to get. If there are multiple
            snapshots with the same name, then the user will be prompted to
            choose a snapshot.
    """
    # Find and validate the current project ID.
    project_id = get_project_id(project_dir)
    resp = send_json_request("snapshot_list", {
        "project_id": project_id,
        "snapshot_uuid": snapshot_uuid,
        })
    snapshots = resp["snapshots"]
    if len(snapshots) == 0:
        raise ValueError("No snapshots found with name {}".format(snapshot_uuid))
    snapshot_idx = 0
    if len(snapshots) > 1:
        print("More than one snapshot found with UUID {}. Which do you want to use?".format(snapshot_uuid))
        for i, snapshot in enumerate(snapshots):
            print("{}. {}".format(i + 1, snapshot["uuid"]))
        snapshot_idx = click.prompt('Please enter a snapshot number from 1 to {}'.format(len(snapshots)), type=int)
        snapshot_idx -= 1
        if snapshot_idx < 0 or snapshot_idx > len(snapshots):
            raise ValueError("Snapshot index {} is out of range".format(snapshot_idx))
    return snapshots[snapshot_idx]["uuid"]


def download_snapshot(snapshot_uuid, target_directory=None, apply_patch=False):
    """Download a snapshot to a local target directory.

    This will recreate the original directory structure of the snapshot. Only
    input files (those found in the project directory during the snapshot
    creation) will be downloaded, not output files specified in the
    project.yaml. The git commit patch will also be downloaded, if any.

    Args:
        snapshot_uuid: The UUID of the snapshot to download.
        target_directory: Directory this snapshot gets downloaded to. If None, the
            snapshot will be downloaded to a temporary directory.
        apply_patch: If True, apply the git patch.

    Returns:
        tuple(str, str, str): The temporary local directory where the snapshot
        has been downloaded, the commit hash to check out if one was included
        in the snapshot, and the git patch to apply if one was included.
    """
    resp = send_json_request("snapshot_get", {
        "snapshot_uuid": snapshot_uuid,
        })

    snapshot_to_s3  = resp["snapshot_to_s3"]
    commit_hash = resp["commit_hash"]
    commit_patch_source = resp["commit_patch_location"]
    input_directories = resp["input_directories"]
    input_files = resp["input_files"]

    if target_directory:
        snapshot_directory = target_directory
    else:
        snapshot_directory = TEMP_SNAPSHOT_DIRECTORY.format(time.time())

    for dir in input_directories:
        target = os.path.join(snapshot_directory, dir)
        os.makedirs(target)

    for source, target in input_files.items():
        target = os.path.join(snapshot_directory, target)
        dirname = os.path.dirname(target)
        os.makedirs(dirname, exist_ok=True)
        copy_file(snapshot_to_s3, source, target, download=True)

    if commit_patch_source is not None:
        commit_patch = COMMIT_PATCH_BASENAME
        target = os.path.join(snapshot_directory, commit_patch)
        copy_file(snapshot_to_s3, commit_patch_source, target, download=True)
    else:
        commit_patch = None

    if apply_patch:
        if commit_hash:
            # Point the repo to the commit associated with the snapshot.
            subprocess.check_call(
                "git reset && git checkout . && git checkout {} && git clean -fxd".format(commit_hash),
                shell=True)

        # Apply the git diff to the session's project directory, if any.
        if commit_patch:
            # Remove the patch after applying.
            subprocess.check_call(
                "git apply {} && rm {}".format(commit_patch, commit_patch),
                shell=True)

    return snapshot_directory, commit_hash, commit_patch



import boto3
import argparse
import click
import datetime
import fnmatch
import json
import jsonschema
import logging
import os
import shutil
import sys
import tabulate

from ray.projects import ProjectDefinition
import ray.projects.scripts as ray_scripts
import ray.ray_constants

import anyscale.conf
from anyscale.project import (PROJECT_ID_BASENAME, get_project_id)
from anyscale.snapshot import (create_snapshot, delete_snapshot,
        describe_snapshot, download_snapshot, get_snapshot_uuid, list_snapshots)
from anyscale.util import (confirm, deserialize_datetime,
        humanize_timestamp, send_json_request)


logging.basicConfig(format=ray.ray_constants.LOGGER_FORMAT)
logger = logging.getLogger(__file__)

if anyscale.conf.AWS_PROFILE is not None:
    logger.info("Using AWS profile %s", anyscale.conf.AWS_PROFILE)
    os.environ["AWS_PROFILE"] = anyscale.conf.AWS_PROFILE


def get_or_create_snapshot(snapshot_uuid, project_definition, yes, local):
    # If no snapshot was provided, create a snapshot.
    if snapshot_uuid is None:
        confirm("No snapshot specified for the command. Create a new snapshot?", yes)
        # TODO: Give the snapshot a name and description that includes this
        # session's name.
        snapshot_uuid = create_snapshot(project_definition, yes, local=True)
    else:
        snapshot_uuid = get_snapshot_uuid(project_definition.root, snapshot_uuid)
    return snapshot_uuid


def load_project_or_throw():
    try:
        project = ray.projects.ProjectDefinition(os.getcwd())
    except (jsonschema.exceptions.ValidationError, ValueError) as e:
        raise click.ClickException(e)

    dir = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(dir, "anyscale_schema.json")) as f:
        schema = json.load(f)

    # Validate the project file.
    try:
        jsonschema.validate(instance=project.config, schema=schema)
    except (jsonschema.exceptions.ValidationError, ValueError) as e:
        raise click.ClickException(e)

    return project


@click.group()
def cli():
    pass


@click.group(
    "project", help="Commands for working with projects.")
def project_cli():
    pass


@click.group(
    "session", help="Commands for working with sessions.")
def session_cli():
    pass


@click.group(
    "snapshot", help="Commands for working with snapshot.")
def snapshot_cli():
    pass


@project_cli.command(
    name="validate",
    help="Validate current project specification.")
@click.option(
    "--verbose", help="If set, print the validated file", is_flag=True)
def project_validate(verbose):
    project = load_project_or_throw()
    print("Project files validated!", file=sys.stderr)
    if verbose:
        print(project.config)


@project_cli.command(
    name="create",
    help="Create a new project within current directory. If the project name "
         "is not provided, this will register the existing project.")
@click.argument("project_name", required=False)
@click.option(
    "--cluster-yaml",
    help="Path to autoscaler yaml. Created by default",
    default=None)
@click.option(
    "--requirements",
    help="Path to requirements.txt. Created by default",
    default=None)
@click.pass_context
def project_create(ctx, project_name, cluster_yaml, requirements):
    project_id_path = os.path.join(ray_scripts.PROJECT_DIR, PROJECT_ID_BASENAME)

    if project_name:
        # Call the normal `ray project create` command.
        ctx.forward(ray_scripts.create)
    elif not os.path.exists(ray_scripts.PROJECT_DIR):
        raise click.ClickException("No registered project found. Did you "
                "forget to pass the project name to `any project create`?")

    try:
        project_definition = load_project_or_throw()
    except click.ClickException as e:
        if not project_name:
            raise click.ClickException("No registered project found. Did you "
                    "forget to pass the project name to `any project create`?")
        else:
            raise e

    project_name = project_definition.config["name"]
    # Add a description of the output_files parameter to the project yaml.
    if not "output_files" in project_definition.config:
        with open(ray_scripts.PROJECT_YAML, 'a') as f:
            f.write("\n".join([
                "",
                "# Pathnames for files and directories that should be saved in a snapshot but",
                "# that should not be synced with a session. Pathnames can be relative to the",
                "# project directory or absolute. Generally, this should be files that were",
                "# created by an active session, such as application checkpoints and logs.",
                "output_files: [",
                "  # For example, uncomment this to save the logs from the last ray job.",
                "  # \"/tmp/ray/session_latest\",",
                "]",
                ]))

    if os.path.exists(project_id_path):
        with open(project_id_path, 'r') as f:
            project_id = int(f.read())
        resp = send_json_request("project_list", {
            "project_id": project_id,
            })
        if len(resp["projects"]) == 0:
            raise click.ClickException("This project has already been "
                    "registered, but its database entry has been deleted?")
        elif len(resp["projects"]) > 1:
            raise click.ClickException("Multiple projects found with the same ID.")

        project = resp["projects"][0]
        if project_name != project["name"]:
            raise ValueError("Project name {} does not match saved project name "
                    "{}".format(project_name, project["name"]))

        raise click.ClickException("This project has already been registered")

    # Add a local database entry for the new Project.
    resp = send_json_request("project_create", {
        "project_name": project_name,
        }, post=True)
    project_id = resp["project_id"]
    with open(project_id_path, 'w+') as f:
        f.write(str(project_id))


@project_cli.command(name="list", help="List all projects currently registered.")
@click.pass_context
def project_list(ctx):
    resp = send_json_request("project_list", {})
    print("Projects:")
    for project in resp["projects"]:
        print(project)


@snapshot_cli.command(name="create", help="Create a snapshot of the current project.")
@click.option(
    "--description",
    help="A description of the snapshot",
    default=None)
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help="Don't ask for confirmation.")
def snapshot_create(description, yes):
    project_definition = load_project_or_throw()
    try:
        create_snapshot(project_definition, yes,
                description=description, local=True)
    except click.Abort as e:
        raise e
    except Exception as e:
        # Creating a snapshot can fail if the project is not found or if some
        # files cannot be copied (e.g., due to permissions).
        raise click.ClickException(e)

@snapshot_cli.command(name="delete", help="Delete a snapshot of the current project with the given UUID.")
@click.argument("uuid")
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help="Don't ask for confirmation.")
def snapshot_delete(uuid, yes):
    project_definition = load_project_or_throw()
    try:
        delete_snapshot(project_definition.root, uuid, yes)
    except click.Abort as e:
        raise e
    except Exception as e:
        # Deleting a snapshot can fail if the project is not found.
        raise click.ClickException(e)


@snapshot_cli.command(name="list", help="List all snapshots of the current project.")
def snapshot_list():
    project_definition = load_project_or_throw()

    try:
        snapshots = list_snapshots(project_definition.root)
    except Exception as e:
        # Listing snapshots can fail if the project is not found.
        raise click.ClickException(e)

    if len(snapshots) == 0:
        print("No snapshots found.")
    else:
        print("Project snaphots:")
        for snapshot in snapshots:
            print(" {}".format(snapshot))

@snapshot_cli.command(name="describe", help="Describe metadata and files of a snapshot.")
@click.argument("name")
def snapshot_describe(name):
    try:
        snapshot = describe_snapshot(name)
    except Exception as e:
        # Describing a snapshot can fail if the snapshot does not exist.
        raise click.ClickException(e)
    print(snapshot)

@snapshot_cli.command(name="download", help="Download a snapshot.")
@click.argument("name")
@click.option("--target-directory", help="Directory this snapshot is downloaded to.")
def snapshot_download(name, target_directory):
    try:
        snapshot = describe_snapshot(name)
    except Exception as e:
        # The snapshot may not exist.
        raise click.ClickException(e)

    if not target_directory:
        target_directory = os.path.join(os.getcwd(), name)

    download_snapshot(snapshot["uuid"], target_directory=target_directory, apply_patch=True)

@session_cli.command(
    name="attach",
    help="Open a console for the given session.")
@click.option(
    "--name", help="Name of the session to open a console for.",
    default=None
)
@click.option("--tmux", help="Attach console to tmux.", is_flag=True)
@click.option("--screen", help="Attach console to screen.", is_flag=True)
def session_attach(name, tmux, screen):
    project_definition = load_project_or_throw()
    project_id = get_project_id(os.getcwd())
    resp = send_json_request("project_sessions", {
        "project_id": project_id,
        "session_name": name,
        })
    sessions = resp["sessions"]

    if len(sessions) != 1:
        raise click.ClickException("Multiple ({}) sessions found with name {}".format(len(sessions), name))

    session = sessions[0]
    ray.autoscaler.commands.attach_cluster(
        project_definition.cluster_yaml(),
        start=False,
        use_tmux=tmux,
        use_screen=screen,
        override_cluster_name=session["name"],
        new=False,
    )


@session_cli.command(
    name="commands",
    help="Print available commands for sessions of this project.")
def session_commands():
    project_definition = load_project_or_throw()
    print("Active project: " + project_definition.config["name"])
    print()

    commands = project_definition.config["commands"]

    for command in commands:
        print("Command \"{}\":".format(command["name"]))
        parser = argparse.ArgumentParser(command["name"], description=command.get("help"), add_help=False)
        params = command.get("params", [])
        for param in params:
            name = param.pop("name")
            if "type" in param: param.pop("type")
            parser.add_argument("--" + name, **param)
        help_string = parser.format_help()
        # Indent the help message by two spaces and print it.
        print("\n".join(["  " + line for line in help_string.split("\n")]))

@session_cli.command(
    name="start",
    context_settings=dict(ignore_unknown_options=True, ),
    help="Start a session based on the current project configuration.")
@click.argument("command", required=False)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.option(
    "--shell", help="If set, run the command as a raw shell command instead of looking up the command in the project.yaml.", is_flag=True)
@click.option(
    "--snapshot", help="If set, start the session from the given snapshot.", default=None)
@click.option(
    "--name", help="A name to tag the session with.", default=None)
def session_start(command, args, shell, snapshot, name):
    project_definition = load_project_or_throw()
    if not name:
        name = project_definition.config["name"]
    # Get the actual command to run. This also validates the command,
    # which should be done before the cluster is started.
    try:
        command, parsed_args, config = project_definition.get_command_info(
            command, args, shell, wildcards=True)
    except ValueError as e:
        raise click.ClickException(e)
    session_runs = ray_scripts.get_session_runs(name, command, parsed_args)

    if len(session_runs) > 1 and not config.get("tmux", False):
        logging.info("Using wildcards with tmux = False would not create "
                     "sessions in parallel, so we are overriding it with "
                     "tmux = True.")
        config["tmux"] = True

    project_id = get_project_id(os.getcwd())
    snapshot_uuid = get_or_create_snapshot(snapshot, project_definition, True, local=True)

    for run in session_runs:
        session_name = run["name"]
        resp = send_json_request("session_list", {
            "project_id": project_id,
            "session_name": session_name,
            })
        if len(resp["sessions"]) == 0:
            resp = send_json_request("session_create", {
                "project_id": project_id,
                "session_name": session_name,
                "snapshot_uuid": snapshot_uuid
                }, post=True)
            session_id = resp["id"]
        else:
            raise click.ClickException("Session with name {} already exists".format(session_name))

@session_cli.command(
    name="sync",
    help="Synchronize a session with a snapshot.")
@click.option(
    "--snapshot", help="The snapshot the session should be synchronized with.", default=None)
@click.option(
    "--name", help="The name of the session to synchronize.", default=None)
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help="Don't ask for confirmation. Confirmation is needed when no snapshot name is provided.")
def session_sync(snapshot, name, yes):
    project_definition = load_project_or_throw()
    project_id = get_project_id(os.getcwd())
    resp = send_json_request("project_sessions", {
        "project_id": project_id,
        "session_name": name,
        })
    sessions = resp["sessions"]

    if len(sessions) == 0:
        raise click.ClickException("No active session matching pattern {} found".format(name))

    # Sync the snapshot to all matching sessions.
    for session in sessions:
        send_json_request("session_sync", {"session_id": session["id"], "snapshot_id": snapshot}, post=True)


@session_cli.command(
    name="execute",
    context_settings=dict(ignore_unknown_options=True, ),
    help="Execute a command in a session.")
@click.argument("command", required=False)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.option(
    "--shell", help="If set, run the command as a raw shell command instead of looking up the command in the project.yaml.", is_flag=True)
@click.option(
    "--name", help="Name of the session to run this command on",
    default=None
)
def session_execute(command, args, shell, name):
    project_definition = load_project_or_throw()

    project_id = get_project_id(os.getcwd())
    resp = send_json_request("project_sessions", {
        "project_id": project_id,
        "session_name": name,
        })
    sessions = resp["sessions"]

    if len(sessions) == 0:
        raise click.ClickException("No active session matching pattern {} found".format(name))

    for session in sessions:
        resp = send_json_request("session_execute", {
            "session_id": session["id"],
            "command": command,
            "params": {},
            "shell": shell,
            }, post=True)


@session_cli.command(name="logs", help="Show logs for the current session.")
@click.option(
    "--name", help="Name of the session to run this command on",
    default=None
)
@click.option(
    "--command-id", help="ID of the command to get logs for",
    default=None
)
def session_logs(name, command_id):
    project_id = get_project_id(os.getcwd())
    # If the command_id is not specified, determine it by getting the last run command from the
    # active session.
    if not command_id:
        resp = send_json_request("project_sessions", {
            "project_id": project_id,
            "session_name": name,
        })
        sessions = resp["sessions"]
        if len(sessions) == 0:
            raise click.ClickException("No active session matching pattern {} found".format(name))
        if len(sessions) > 1:
            raise click.ClickException("Multiple active sessions matching pattern {} found".format(name))
        session = sessions[0]
        resp = send_json_request("session_describe", {
            "session_id": session["id"],
        })
        # Search for latest run command
        last_created_at = datetime.datetime.min
        for command in resp["commands"]:
            created_at = deserialize_datetime(command["created_at"])
            if created_at > last_created_at:
                last_created_at = created_at
                command_id = command["session_command_id"]
        if not command_id:
            raise click.ClickException("No comand was run yet on the latest active session {}".format(session["name"]))
    resp = send_json_request("session_execution_logs", {
        "session_command_id": command_id,
        "out_from_bytes": 0,
        "err_from_bytes": 0
    })
    # TODO(pcm): We should have more options here in the future
    # (e.g. show only stdout or stderr, show only the tail, etc).
    print("stdout:")
    print(resp["out"])
    print("stderr:")
    print(resp["err"])


@session_cli.command(name="stop", help="Stop the current session.")
@click.option(
    "--name", help="Name of the session to stop",
    default=None
)
@click.pass_context
def session_stop(ctx, name):
    project_id = get_project_id(os.getcwd())
    resp = send_json_request("project_sessions", {
        "project_id": project_id,
        "session_name": name,
        })
    sessions = resp["sessions"]
    if len(sessions) == 0:
        raise click.ClickException("No active session matching pattern {} found".format(name))

    for session in sessions:
        # Stop the session and mark it as stopped in the database.
        send_json_request("session_stop", {
            "session_id": session["id"],
            }, post=True)

@session_cli.command(name="list", help="List all sessions of the current project.")
@click.option(
    "--name",
    help="Name of the session. If provided, this prints the snapshots that were applied and commands that ran for all sessions that match this name.",
    default=None)
@click.option(
    "--all", help="List all sessions, including inactive ones.", is_flag=True)
def session_list(name, all):
    project_id = get_project_id(os.getcwd())
    project_definition = load_project_or_throw()
    print("Active project: " + project_definition.config["name"])

    resp = send_json_request("session_list", {
        "project_id": project_id,
        "session_name": name,
        "active_only": not all,
        })
    sessions = resp["sessions"]

    if name is None:
        print()
        table = []
        for session in sessions:
            created_at = humanize_timestamp(deserialize_datetime(session["created_at"]))
            record = [session["name"], created_at, session["latest_snapshot_uuid"]]
            if all:
                table.append([" Y" if session["active"] else " N"] + record)
            else:
                table.append(record)
        if not all:
            print(tabulate.tabulate(table, headers=["SESSION", "CREATED", "SNAPSHOT"], tablefmt="plain"))
        else:
            print(tabulate.tabulate(table, headers=["ACTIVE", "SESSION", "CREATED", "SNAPSHOT"], tablefmt="plain"))
    else:
        sessions = [session for session in sessions if session["name"] == name]
        for session in sessions:
            resp = send_json_request("session_describe", {
                "session_id": session["id"],
                })

            print()
            snapshot_table = []
            for applied_snapshot in resp["applied_snapshots"]:
                snapshot_uuid = applied_snapshot["snapshot_uuid"]
                created_at = humanize_timestamp(deserialize_datetime(applied_snapshot["created_at"]))
                snapshot_table.append([snapshot_uuid, created_at])
            print(tabulate.tabulate(snapshot_table, headers=["SNAPSHOT applied to {}".format(session["name"]), "APPLIED"], tablefmt="plain"))

            print()
            command_table = []
            for command in resp["commands"]:
                created_at = humanize_timestamp(deserialize_datetime(command["created_at"]))
                command_table.append([command["command"], command["session_command_id"], created_at])
            print(tabulate.tabulate(command_table, headers=["COMMAND run in {}".format(session["name"]), "ID", "CREATED"], tablefmt="plain"))


cli.add_command(project_cli)
cli.add_command(session_cli)
cli.add_command(snapshot_cli)

def main():
    return cli()

if __name__ == '__main__':
    main()
